import argparse
import asyncio
import importlib
import json
import logging
import typing
from .artifacts import ArtifactFactory, gather_artifacts
from .transformations import cancel_all_transformations, Transformation


LOGGER = logging.getLogger("beaver")


class Formatter(logging.Formatter):
    # Based on https://stackoverflow.com/a/56944256/1150961.
    cyan = "\x1b[36;20m"
    green = "\x1b[32;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    colors_by_level = {
        'debug': cyan,
        'info': green,
        'warning': yellow,
        'error': red,
        'critical': bold_red
    }

    def format(self, record: logging.LogRecord) -> str:
        record._prefix = self.colors_by_level[record.levelname.lower()]
        record._suffix = self.reset
        return super().format(record)


def __main__(args: typing.Iterable[str] = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--digest", "-d", help="file containing composite artifact digests",
                        default=".beaverdigests")
    parser.add_argument("--file", "-f", help="file containing artifact and transform definitions",
                        default="beaver.py")
    parser.add_argument("--num_concurrent", "-c", help="number of concurrent transformations",
                        type=int, default=1)
    parser.add_argument("--dry-run", "-n", help="print transformations without executing them",
                        action="store_true")
    parser.add_argument("--log_level", "-l", help="level of log messages to emit", default="info",
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], type=str.upper)
    parser.add_argument("artifacts", help="artifacts to generate", nargs="+")
    args = parser.parse_args(args)

    # Configure logging to stderr.
    root_logger = logging.getLogger()
    root_logger.setLevel(args.log_level)
    handler = logging.StreamHandler()
    handler.setFormatter(Formatter("\U0001f9ab %(_prefix)s%(levelname)s%(_suffix)s: %(message)s"))
    root_logger.addHandler(handler)

    Transformation.DRY_RUN = args.dry_run

    # Load the artifact and transformation configuration.
    try:
        spec = importlib.util.spec_from_file_location("config", args.file)
        config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(config)
        LOGGER.debug("loaded %d artifacts from `%s`", len(ArtifactFactory.REGISTRY), args.file)
    except FileNotFoundError:
        LOGGER.error("beaver configuration cannot be loaded from %s", args.file)
        return 1

    # Load the composite digests but only retain the composite digests of known artifacts.
    try:
        with open(args.digest) as fp:
            composite_digests = json.load(fp)
        composite_digests = {name: digest for name, digest in composite_digests.items()
                             if name in ArtifactFactory.REGISTRY}
        Transformation.COMPOSITE_DIGESTS = composite_digests
        LOGGER.debug("loaded %d composite digests from `%s`", len(composite_digests), args.digest)
    except FileNotFoundError:
        LOGGER.debug("did not load composite digests because the file `%s` does not exist",
                     args.digest)
        pass

    try:
        # Get the targets we want to build and wait for them to complete.
        artifacts = [ArtifactFactory.REGISTRY[name] for name in args.artifacts]
        asyncio.run(gather_artifacts(*artifacts, num_concurrent=args.num_concurrent))
    finally:
        # Save the updated composite digests.
        with open(args.digest, "w") as fp:
            json.dump(Transformation.COMPOSITE_DIGESTS, fp, indent=4)
        LOGGER.debug("saved %d composite digests to `%s`", len(Transformation.COMPOSITE_DIGESTS),
                     args.digest)
        cancel_all_transformations()


if __name__ == "__main__":  # pragma: no cover
    __main__()



import asyncio
import contextlib
import glob
import hashlib
import logging
import os
import re
import shlex
import typing
from . import transformations


LOGGER = logging.getLogger(__name__)


class ArtifactFactory(type):
    """
    Metaclass that creates :class:`Artifact` instances or returns existing instances.

    Attributes:
        REGISTRY: Mapping of artifact names to :class:`Artifact` instances. The registry is used to
            keep track of all artifacts and ensure artifact names are unique.
    """
    # Declare using an old-school type hint because Sphinx struggles with circular references.
    REGISTRY = {}  # type: dict[str, Artifact]

    def __init__(self, name, bases, members):
        super(ArtifactFactory, self).__init__(name, bases, members)

    def __call__(self, name, *args, **kwargs):
        # Evaluate the fully qualified name in light of the current group stack.
        name = Group.evaluate_qualified_name(name)
        # Try to retrieve the instance from the registry.
        if instance := self.REGISTRY.get(name):
            if (cls := instance.__class__) is not self:
                raise ValueError(f"cannot create instance of {self} because artifact `{name}` with "
                                 f"type {cls} already exists")
            return instance
        # Create a new instance, register it with the current group if any, and add it to the
        # registry.
        instance = super(ArtifactFactory, self).__call__(name, *args, **kwargs)
        Group.append(instance)
        self.REGISTRY[name] = instance
        return instance


class Artifact(metaclass=ArtifactFactory):
    """
    Artifact generated by a transformation.

    Args:
        name: Unique name of the artifact.

    Raises:
        ValueError: If the given :code:`name` is already in use by another artifact.

    Attributes:
        children: Transformations that consume this artifact; empty if the artifact is a leaf of the
            directed acyclic graph.
        parent: Transformation that generates this artifact; :code:`None` if the artifact is a root
            of the directed acyclic graph.
        digest: Concise summary of the artifact; :code:`None` if the artifact does not exist, cannot
            be summarized, or should always be generated using its :attr:`parent` transform.
    """
    def __init__(self, name: str) -> None:
        self.name = name
        self._parent = None
        self.children = []

    children: typing.Iterable[transformations.Transformation]

    @property
    def parent(self) -> 'transformations.Transformation':
        return self._parent

    @parent.setter
    def parent(self, value: 'transformations.Transformation') -> None:
        if self._parent is not None:
            raise RuntimeError(f"{self} is already associated with a transformation")
        self._parent = value

    @property
    def digest(self) -> bytes:
        return None

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    async def __call__(self):
        if self.parent is None:
            return
        return await self.parent

    def __await__(self):
        # See https://stackoverflow.com/a/57078217/1150961 for details.
        return (yield from self().__await__())


class Group(Artifact):
    """
    Artifact representing a group of other artifacts.

    Args:
        name: Name of the group. The group name is added as a prefix for all artifacts created
            within the context manager of the group.

    Attributes:
        members: Members of the group.
    """
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.members = []

    def __enter__(self) -> "Group":
        if self in self.STACK:  # pragma: no cover
            raise RuntimeError(f"{self} is already in the stack")
        self.STACK.append(self)
        return self

    def __exit__(self, *_):
        if (other := self.STACK.pop()) is not self:  # pragma: no cover
            raise RuntimeError(f"expected the last element to be {self} but got {other}")

    async def __call__(self):
        await asyncio.gather(*self.members)

    STACK: list["Group"] = []

    @classmethod
    def evaluate_qualified_name(cls, name: str) -> str:
        """
        Evaluate the fully qualified name of an artifact given its context.

        Args:
            name: Unqualified name.

        Returns:
            name: Qualified name.
        """
        if not cls.STACK:
            return name
        return os.path.join(cls.STACK[-1].name, name)

    @classmethod
    def append(cls, artifact: Artifact) -> None:
        """
        Append an artifact to the current group if available.

        Args:
            artifact: Artifact to append to the current group.
        """
        if not cls.STACK:
            return
        cls.STACK[-1].members.append(artifact)


@contextlib.contextmanager
def group_artifacts(*names: str, squeeze=True) -> list[Group]:
    """
    Group artifacts in a context.

    Args:
        names: Name of nested groups.
        squeeze: Whether to return a single element if there is only one group.

    Returns:
        groups: Artifacts representing the nested groups.
    """
    groups = []
    try:
        for name in names:
            groups.append(Group(name).__enter__())
        yield groups[0] if len(groups) == 1 and squeeze else groups
    except Exception:
        Group.STACK.clear()
        raise
    finally:
        [group.__exit__() for group in reversed(groups)]


class File(Artifact):
    """
    Artifact representing a file.
    """
    def __init__(self, name: str) -> None:
        super().__init__(name)
        if re.search(r"\s", self.name):
            LOGGER.warning("whitespace in `%s` is a recipe for disaster; expect the unexpected",
                           self.name)
        self._last_modified = None
        self._digest = None

    async def __call__(self):
        # Omit directory creation and existence checks during dry run.
        if transformations.Transformation.DRY_RUN:
            await super().__call__()
            return

        # Create the parent directory if necessary.
        if dirname := os.path.dirname(self.name):
            os.makedirs(dirname, exist_ok=True)
        # Generate the artifact and verify the file exists.
        await super().__call__()
        if not os.path.exists(self.name):
            if self._parent:
                message = f"{self._parent} did not generate {self}"
            else:
                message = f"{self} does not exist and cannot be generated"
            raise FileNotFoundError(message)

    @property
    def digest(self):
        try:
            # Only re-evaluate the digest if the file has not been modified since we last computed
            # the hash.
            last_modified = os.stat(self.name).st_mtime
            if self._last_modified is None or self._last_modified < last_modified:
                algorithm = hashlib.sha256()
                with open(self.name, 'rb') as fp:
                    while (chunk := fp.read(4096)):
                        algorithm.update(chunk)
                self._digest = algorithm.digest()
                self._last_modified = last_modified
            return self._digest
        except FileNotFoundError:
            return None

    @classmethod
    def glob(self, pattern: str) -> typing.Iterable["File"]:
        """
        Create a list of file artifacts based on a glob pattern.

        Args:
            pattern: Pattern passed to :func:`glob.glob`.

        Returns:
            artifacts: File artifacts.
        """
        return normalize_artifacts([filename for filename in glob.glob(pattern)])

    def __str__(self):
        return shlex.quote(self.name)


def normalize_artifacts(
    artifacts: typing.Union[str, Artifact, typing.Iterable[typing.Union[str, Artifact]]]) \
        -> typing.Iterable[Artifact]:
    r"""
    Normalize one or more artifacts. Strings are implicitly converted to :class:`File`\ s.

    Args:
        artifacts: One or more artifacts or transformations.

    Returns:
        normalized: Normalized artifacts.
    """
    if artifacts is None:
        return []
    if not isinstance(artifacts, typing.Iterable) or isinstance(artifacts, str):
        artifacts = [artifacts]
    normalized = []
    for artifact in artifacts:
        if isinstance(artifact, str):
            normalized.append(File(artifact))
        elif isinstance(artifact, Artifact):
            normalized.append(artifact)
        elif isinstance(artifact, transformations.Transformation):
            normalized.extend(artifact.outputs)
        else:
            raise TypeError(artifact)
    return normalized


async def gather_artifacts(*artifacts_or_transformations, num_concurrent: int = None) \
       -> typing.Coroutine:
    """
    Gather one or more artifacts and wrap them in a coroutine.

    Args:
        *artifacts_or_transformations: Artifacts to gather and/or transformations whose artifacts to
            gather.
        num_concurrent: Maximum number of concurrent transformations. Defaults to unrestricted.

    Returns:
        gathered: Coroutine that awaits all gathered artifacts.
    """
    try:
        if num_concurrent is not None:
            transformations.Transformation.SEMAPHORE = asyncio.Semaphore(num_concurrent)
        gathered = []
        for item in artifacts_or_transformations:
            if isinstance(item, Artifact):
                gathered.append(item)
            elif isinstance(item, transformations.Transformation):
                gathered.extend(item)
            else:
                raise TypeError(item)
        await asyncio.gather(*gathered)
    finally:
        transformations.Transformation.SEMAPHORE = None



from .artifacts import ArtifactFactory, Group
from .transformations import Shell, Transformation
from .artifacts import *  # noqa: F401, F403
from .transformations import *   # noqa: F401, F403
from .util import *  # noqa: F401, F403


def reset() -> None:
    ArtifactFactory.REGISTRY.clear()
    Group.STACK.clear()
    Shell.ENV.clear()
    Transformation.COMPOSITE_DIGESTS.clear()
    Transformation.DRY_RUN = False
    Transformation.SEMAPHORE = None



import contextlib
import os


@contextlib.contextmanager
def working_directory(directory):
    """
    Temporarily change the working directory using a context manager.
    """
    current = os.getcwd()
    os.chdir(directory)
    yield directory
    os.chdir(current)



import aiohttp
import asyncio
import hashlib
import logging
import os
import re
import sys
import typing
from . import artifacts


LOGGER = logging.getLogger(__name__)


def cancel_all_transformations() -> None:
    """
    Cancel all running transformations.
    """
    try:
        tasks = asyncio.all_tasks()
    except RuntimeError:
        return

    for task in tasks:
        task.cancel()


class Transformation:
    """
    Base class for transformations that generates outputs given inputs. Inheriting classes should
    implement :meth:`execute`.

    Args:
        outputs: Artifacts to generate.
        inputs: Artifacts consumed by the transformation.

    Attributes:
        COMPOSITE_DIGESTS: Mapping of artifact names to composite digests. See
            :func:`evaluate_composite_digests` for details.
        SEMAPHORE: Semaphore to limit the number of concurrent transformations being executed.
        DRY_RUN: Whether to show transformations without executing them.
    """
    def __init__(self, outputs: typing.Iterable["artifacts.Artifact"],
                 inputs: typing.Iterable["artifacts.Artifact"]) -> None:
        self.inputs = artifacts.normalize_artifacts(inputs)
        for input in self.inputs:
            input.children.append(self)
        self.outputs = artifacts.normalize_artifacts(outputs)
        for output in self.outputs:
            output.parent = self
        self.future = None

    def __iter__(self):
        for output in self.outputs:
            yield output

    def evaluate_composite_digests(self) -> dict[str, bytes]:
        """
        Evaluate composite digests for all :code:`outputs` of the transformation.

        A composite digest is defined as the digest of the digests of inputs and the digest of the
        output, i.e. it is a joint, concise summary of both inputs and outputs. If any composite
        digest differs from the digest in :attr:`COMPOSITE_DIGESTS` or is :code:`None`, the
        transformation needs to be executed. A composite digest is :code:`None` if the corresponding
        output digest is :code:`None` or any input digests are :code:`None`.

        Returns:
            digests: Mapping from output names to composite digests.
        """
        # Initialise empty digests for the outputs.
        digests = {o.name: None for o in self.outputs}

        # Evaluate input digests and abort if any of them are missing.
        input_hasher = hashlib.sha256()
        for i in self.inputs:
            if i.digest is None:
                return digests
            input_hasher.update(i.digest)

        # Construct composite digests for the outputs.
        for output in self.outputs:
            if output.digest is None:
                continue
            hasher = input_hasher.copy()
            hasher.update(output.digest)
            digests[output.name] = hasher.hexdigest()
        return digests

    async def __call__(self) -> None:
        # Wait for all inputs artifacts.
        await asyncio.gather(*self.inputs)
        # Create a future if required and wait for it to complete.
        if not self.future:
            self.future = asyncio.create_task(self._execute())
        await self.future

    async def _execute(self) -> None:
        """
        Private wrapper to determine whether the transformation needs to be executed and wrap it in
        a semaphore to limit concurrency.
        """
        # Check if any composite indices have changed. If no, there's nothing further to be done.
        composite_digests = self.evaluate_composite_digests()
        stale_artifacts = [name for name, digest in composite_digests.items() if digest is None or
                           digest != self.COMPOSITE_DIGESTS.get(name)]
        if not stale_artifacts:
            LOGGER.debug("composite digests %s are unchanged; %s will not be executed",
                         composite_digests, self)
            LOGGER.info("\U0001f7e2 artifacts [%s] are up to date",
                        ", ".join(o.name for o in self.outputs))
            return
        LOGGER.debug("composite digests of %s have changed: %s", ", ".join(stale_artifacts),
                     composite_digests)

        if self.DRY_RUN:
            LOGGER.info("\U0001f7e1 artifacts [%s] are stale; transformation not scheduled because "
                        "of dry run", ", ".join(stale_artifacts))
            return

        LOGGER.info("\U0001f7e1 artifacts [%s] are stale; schedule transformation",
                    ", ".join(stale_artifacts))

        try:
            if self.SEMAPHORE is None:
                await self.execute()
            else:
                async with self.SEMAPHORE:
                    await self.execute()

            # Update the composite digests.
            composite_digests = self.evaluate_composite_digests()
            LOGGER.info("\u2705 generated artifacts [%s]", ", ".join(o.name for o in self.outputs))
            self.COMPOSITE_DIGESTS.update(composite_digests)
        except Exception as ex:
            LOGGER.error("\u274c failed to generate artifacts [%s]: %s",
                         ", ".join(o.name for o in self.outputs), ex)
            raise

    async def execute(self) -> None:
        """
        Execute the transformation.
        """
        raise NotImplementedError

    def __repr__(self):
        inputs = ", ".join(map(repr, self.inputs))
        outputs = ", ".join(map(repr, self.outputs))
        return f"{self.__class__.__name__}([{inputs}] -> [{outputs}])"

    def __await__(self):
        # See https://stackoverflow.com/a/57078217/1150961 for details.
        return (yield from self().__await__())

    COMPOSITE_DIGESTS: dict[str, bytes] = {}
    SEMAPHORE: typing.Optional[asyncio.Semaphore] = None
    DRY_RUN: bool = False


class _Sleep(Transformation):
    """
    Sleeps for a given number of seconds and create any file artifact outputs. Mostly used for
    testing and debugging.

    Args:
        outputs: Artifacts to generate.
        inputs: Artifacts consumed by the transformation.
        time: Number of seconds to sleep for.
    """
    def __init__(self, outputs: typing.Iterable["artifacts.Artifact"],
                 inputs: typing.Iterable["artifacts.Artifact"], *, time: float) -> None:
        super().__init__(outputs, inputs)
        self.time = time

    async def execute(self) -> None:
        LOGGER.debug("running %s for %f seconds...", self, self.time)
        await asyncio.sleep(self.time)
        for output in self.outputs:
            if isinstance(output, artifacts.File):
                with open(output.name, "w") as fp:
                    fp.write(output.name)
                LOGGER.debug("created %s", output)
        LOGGER.debug("completed %s", self)


class Download(Transformation):
    """
    Download a file and verify its digest.

    Args:
        output: Output artifact for the downloaded data.
        url: Url to download from.
        digest: Expected digest of the downloaded data.
    """
    def __init__(self, output: "artifacts.File", url: str, digest: typing.Union[str, bytes]) \
            -> None:
        super().__init__(output, [])
        self.digest = bytes.fromhex(digest) if isinstance(digest, str) else digest
        self.url = url

    async def execute(self) -> None:
        # Abort if we already have the right file.
        output, = self.outputs
        if output.digest == self.digest:
            return
        # Download the file.
        async with aiohttp.ClientSession() as session:
            async with session.get(self.url) as response:
                with open(output.name, "wb") as fp:
                    fp.write(await response.read())
        if output.digest != self.digest:
            raise ValueError(f"expected digest `{self.digest.hex()}` but got "
                             f"`{output.digest.hex()}` for {output}")


class Subprocess(Transformation):
    r"""
    Execute a subprocess.

    The transformation supports variable substitution using f-strings and Makefile syntax:

    - :code:`$@` represents the first output.
    - :code:`$<` represents the first input.
    - :code:`$^` represents all inputs.
    - :code:`$$@` represents the literal :code:`$@`, i.e. double dollars are properly escaped.
    - :code:`{outputs[0].name}` represents the name of the first output. The available f-string
      variables are :code:`outputs` and :code:`inputs`, each a list of
      :class:`Artifact <.artifacts.Artifact>`\ s.
    - :code:`$!` represents the current python interpreter.

    Environment variables are inherited by default, but global environment variables for all
    :class:`Subprocess` transformations can be specified in :attr:`ENV`, and specific environment
    variables can be specified using the :code:`env` argument. Environment variables that are not
    "truthy" are removed from the environment of the transformation.

    Args:
        outputs: Artifacts to generate.
        inputs: Artifacts consumed by the transformation.
        cmd: Command to execute. Must be a single string if :code:`shell` is truthy and a sequence
            of strings otherwise.
        env: Mapping of environment variables.
        shell: Whether to execute the command through the shell, e.g. to expand the home directory
            :code:`~` or pipe information between processes or files using :code:`|`, :code:`<`, or
            :code:`>`. See :class:`subprocess.Popen` for details, including security considerations.
        **kwargs: Keyword arguments passed to :func:`asyncio.subprocess.create_subprocess_shell` (if
            :code:`shell == True`) or :func:`asyncio.subprocess.create_subprocess_exec` (if
            :code:`shell == False`).

    Raises:
        ValueError: If :code:`shell == True` and :code:`cmd` is not a string, or
            :code:`shell == False` and :code:`cmd` is not a list of strings.

    Attributes:
        ENV: Default mapping of environment variables for all :class:`Subprocess` transformations.
    """
    def __init__(self, outputs: typing.Iterable["artifacts.Artifact"],
                 inputs: typing.Iterable["artifacts.Artifact"],
                 cmd: typing.Union[str, typing.Iterable[str]], *, env: dict[str, str] = None,
                 shell: bool = False, **kwargs) -> None:
        super().__init__(outputs, inputs)
        if shell and not isinstance(cmd, str):
            raise ValueError(f"`cmd` must be a string if `shell == True` but got {cmd}")
        if not shell and not isinstance(cmd, list):
            raise ValueError(f"`cmd` must be a list of strings if `shell == False` but got {cmd}")
        self.cmd = cmd
        self.shell = shell
        self.env = env or {}
        self.kwargs = kwargs

    def _apply_substitutions(self, part: str) -> str:
        part = part.format(outputs=self.outputs, inputs=self.inputs)
        # Apply Makefile-style and python interpreter substitutions.
        rules = {
            r"@": self.outputs[0],
            r"<": self.inputs[0] if self.inputs else None,
            r"\^": " ".join(input.name for input in self.inputs),
            r"!": sys.executable,
        }
        for key, value in rules.items():
            part = re.sub(r"(?<!\$)\$" + key, str(value), part)
        return part

    async def execute(self) -> None:
        # Prepare the command.
        if self.shell:
            cmd = self._apply_substitutions(self.cmd)
        else:
            cmd = [self._apply_substitutions(str(part)) for part in self.cmd]
        LOGGER.info("\u2699\ufe0f execute %s command `%s`", "shell" if self.shell else "subprocess",
                    cmd)
        # Call the process.
        env = os.environ | self.ENV | self.env
        env = {key: str(value) for key, value in env.items() if value}
        if self.shell:
            process = await asyncio.subprocess.create_subprocess_shell(cmd, env=env, **self.kwargs)
        else:
            process = await asyncio.subprocess.create_subprocess_exec(*cmd, env=env, **self.kwargs)
        status = await process.wait()
        if status:
            raise RuntimeError(f"{self} failed with status code {status}")

    ENV: dict[str, str] = {}


class Shell(Subprocess):
    """
    Execute a command in the shell. See :class:`Subprocess` for details.
    """
    def __init__(self, outputs: typing.Iterable["artifacts.Artifact"],
                 inputs: typing.Iterable["artifacts.Artifact"],
                 cmd: str, *, env: dict[str, str] = None, **kwargs) -> None:
        super().__init__(outputs, inputs, cmd, env=env, shell=True, **kwargs)


class Functional(Transformation):
    """
    Apply a python function.

    Args:
        outputs: Artifacts to generate.
        inputs: Artifacts consumed by the transformation.
        func: Function to execute.
        *args: Positional arguments passed to :code:`func`.
        *kwargs: Keyword arguments passed to :code:`func`.
    """
    def __init__(self, outputs: typing.Iterable["artifacts.Artifact"], inputs:
                 typing.Iterable["artifacts.Artifact"], func: typing.Callable, *args, **kwargs) \
            -> None:
        super().__init__(outputs, inputs)
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def execute(self) -> None:
        return self.func(self.outputs, self.inputs, *self.args, **self.kwargs)



import re
from setuptools import find_packages, setup


# Load the README and strip out Sphinx directives that PyPI cannot handle.
with open("README.rst") as fp:
    long_description = fp.read()
replacements = [
    (".. doctest::", ".. code-block:: python"),
    (".. testcode::", ".. code-block:: python"),
    (":math:", ":code:"),
    (":func:", ":code:"),
    (":class:", ":code:"),
    (":meth:", ":code:"),
    (":mod:", ":code:"),
]
for old, new in replacements:
    long_description = long_description.replace(old, new)
long_description = re.sub(
    r"((\.\. automodule:: .*?$)|(\.\. toctree::)|(\.\. plot:: .*?$))", r".. code-block::\n\n  \1",
    long_description, flags=re.MULTILINE
)

# Load the version number.
try:
    with open('VERSION') as fp:
        version = fp.read().strip()
except FileNotFoundError:
    version = 'dev'

setup(
    name="beaver-build",
    packages=find_packages(),
    author="Till Hoffmann",
    author_email="niches_osmosis0a@icloud.com",
    version=version,
    install_requires=[
        "aiohttp",
    ],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    extras_require={
        "tests": [
            "flake8",
            "pytest",
            "pytest-cov",
            "twine",
        ],
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
        ]
    },
    entry_points={
        "console_scripts": [
            "beaver = beaver_build.cli:__main__",
        ]
    },
)
